"""
Created on 23Jan.,2018

@author: dlawrence
"""


# Change this when you introduce breaking changes
PYREFERENCE_JSON_VERSION = 5
PYREFERENCE_JSON_VERSION_KEY = "pyreference_json_version"

# Keys used in dictionary (serialized to JSON)
CHROM = "chr"
START = "start"
END = "stop"
STRAND = "strand"

# Other 
IS_CODING = "is_coding"


BEST_REGION_TYPE_ORDER = ["coding", "5PUTR", "3PUTR", "non coding", "intron"]

